// ignore_for_file: public_member_api_docs

import 'dart:developer';

import 'package:collection/collection.dart';
import 'package:fl_chart/fl_chart.dart';
import 'package:healthrian_ble_support_for_wearecg12/'
    'healthrian_ble_support_for_wearecg12.dart';
import 'package:healthrian_common_support/healthrian_common_support.dart';
import 'package:lzma/lzma.dart';

mixin Ecg {
  /// [maxBufferStreamEcg] decides a maximum length of buffered Stream.
  ///
  /// For 40 seconds buffer and 250 sampling rate,
  /// It uses 40 * 250 by default.
  int maxBufferStreamEcg = 40 * 250;

  /// [lengthBufferStreamEcg] decides a length of buffered Stream,
  /// after flushing buffer when reaching a maximum size.
  ///
  /// For 30 seconds buffer and 250 sampling rate,
  /// It uses 30 * 250 by default.
  int lengthBufferStreamEcg = 30 * 250;

  /// [screenBufferStreamingEcg] decides a length of buffered Stream,
  /// which is shown to screen.
  ///
  /// For 5 seconds buffer and 250 sampling rate,
  /// It uses 5 * 250 by default.
  int screenBufferStreamingEcg = 5 * 250;

  /// [isFilteredEcg] decides whether ecg signal is filtered.
  bool isFilteredEcg = true;

  /// [isCompressedEcg] decides whether ecg signal is compressed.
  ///
  /// If it is `true`, ecg signal is decompressed by LZMA algorithm.
  bool isCompressedEcg = false;

  final _dataFilter = DataFilter(8);

  final _adaptiveHR = AdaptiveHR(0);

  bool isPaused = false;

  /// [skipSampleCount] is the number of samples to be skipped.
  /// For example, [skipSampleCount] is 5 by default,
  /// The plotting stream only emits a 1/5 samples,
  /// which discards 4/5 samples.
  int skipSampleCount = 5;
}

extension BLEEcg on BLEManager {
  /// [startMeasurementEcg] sends read_ecg_packet to device,
  ///  and then, it registers all ecg streams of [streams].
  Future<void> startMeasurementEcg() async {
    await _sendCommandHolterTime();
    await _sendCommandReadEcg();

    if (!isPaused) {
      streams
        ..rawStream = _getStream().broadcast
        ..decompressedStream = _getDecompressedStream().broadcast
        ..recentStream = _getRecentStream().broadcast
        ..leadStateStream = _getLeadStateStream().broadcast
        ..heartRateStream = _getHeartRateStream().broadcast
        ..bufferedStreamEcg = _getBufferedStream().broadcast
        ..bufferedPlotStreamEcg = _getBufferedPlotStream().broadcast
        ..bufferedPlotRTStreamEcg = _getBufferedPlotRTStream().broadcast;
    }

    isPaused = false;
  }

  Future<void> getLeadOffState() async {
    await _sendCommandGetLeadState();
  }

  Future<void> stopLeadOffState() async {
    await _sendCommandStopLeadState();
  }

  /// [pauseMeasurementEcg]
  Future<void> pauseMeasurementEcg() async {
    if (isPaused) {
      await startMeasurementEcg();
      return;
    }
    isPaused = true;
    await stopMeasurementEcg();
  }

  /// [stopMeasurementEcg] sends stop_ecg_packet to device.
  Future<void> stopMeasurementEcg() async {
    final control = points[Characteristics.HEH_CONTROL_POINT];
    if (control == null) return;

    log('Writes a packet ${Commands.packetStopEcg}');
    await control.write(Commands.packetStopEcg);
  }

  Future<void> _sendCommandReadEcg() async {
    final control = points[Characteristics.HEH_CONTROL_POINT];
    if (control == null) return;

    log('Writes a packet ${Commands.packetReadEcg}');
    await control.write(Commands.packetReadEcg);
  }

  Future<void> _sendCommandGetLeadState() async {
    final control = points[Characteristics.HEH_CONTROL_POINT];
    if (control == null) return;

    log('Writes a packet ${Commands.packetGetLeadState}');
    await control.write(Commands.packetGetLeadState);
  }

  Future<void> _sendCommandStopLeadState() async {
    final control = points[Characteristics.HEH_CONTROL_POINT];
    if (control == null) return;

    log('Writes a packet ${Commands.packetStopLeadState}');
    await control.write(Commands.packetStopLeadState);
  }

  Future<void> _sendCommandHolterTime() async {
    final control = points[Characteristics.HEH_CONTROL_POINT];
    if (control == null) return;

    log('Writes a packet ${Commands.packetSetHolterTime}');
    await control.write(Commands.packetSetHolterTime);
  }

  Stream<List<int>> _getStream() async* {
    final measurement = points[Characteristics.HEH_MEASUREMENT];
    if (measurement == null) return;
    await measurement.setNotifyValue(true);
    await for (final res in measurement.value) {
      yield res;
    }
  }

  Stream<List<int>> _getDecompressedStream() async* {
    final events = streams.rawStream;
    if (events == null) return;
    await for (final event in events) {
      yield lzma.decode(event);
    }
  }

  Stream<List<int>>? get _events {
    return isCompressedEcg ? streams.decompressedStream : streams.rawStream;
  }

  Stream<List<double>> _getRecentStream() async* {
    final events = _events;
    if (events == null) return;
    final res = <double>[];
    await for (final event in events) {
      if (event.isEmpty) continue;
      final id = event.elementAtOrNull(3);
      if (id != 0x11) continue;
      final itemCount = event.length ~/ 3;
      res.clear();
      for (var i = 2; i < itemCount; i++) {
        final raw = toDouble(
          event[i * 3 + 0],
          event[i * 3 + 1],
          event[i * 3 + 2],
        );
        // res.add(isFilteredEcg ? _filter(raw) : raw);
        res.add(raw);
      }
      yield res;
    }
  }

  Stream<LeadState> _getLeadStateStream() async* {
    final events = _events;
    if (events == null) return;

    bool checkBit(int value, int bit) => (value & (1 << bit)) == 0;
    final leadState = LeadState.empty();

    // const skipNumberOfPackets = 27;
    // var count = 0;
    var now = DateTime.now();

    await for (final event in events) {
      if (event.isEmpty) continue;
      final id = event.elementAtOrNull(3);
      if (id != 0x11) continue;

      // count++;
      // if (count % skipNumberOfPackets != 0) continue;
      final currentTime = DateTime.now();
      if (currentTime.difference(now) <= const Duration(milliseconds: 500)) {
        continue;
      }

      now = currentTime;

      final leadStateHigh = event.elementAtOrNull(4);
      final leadStateLow = event.elementAtOrNull(5);

      if (leadStateHigh == null || leadStateLow == null) continue;

      yield leadState.copyWith(
        ra: checkBit(leadStateHigh, 1),
        ll: checkBit(leadStateLow, 2),
        la: checkBit(leadStateLow, 1),
        v1: checkBit(leadStateLow, 7),
        v2: checkBit(leadStateLow, 3),
        v3: checkBit(leadStateLow, 4),
        v4: checkBit(leadStateLow, 5),
        v5: checkBit(leadStateLow, 6),
        v6: checkBit(leadStateLow, 0),
      );
    }
  }

  Stream<int> _getHeartRateStream() async* {
    final events = _events;
    if (events == null) return;

    await for (final event in events) {
      if (event.isEmpty) continue;
      final id = event.elementAtOrNull(3);
      if (id != 0x11) continue;

      final leadStateHigh = event.elementAtOrNull(4);
      final leadStateLow = event.elementAtOrNull(5);

      if (leadStateHigh != 0x00 || leadStateLow != 0x00) {
        yield 0;
        continue;
      }

      final itemCount = event.length ~/ 3;

      var count = 0;
      for (var i = 2; i < itemCount; i++) {
        count++;
        if (count % 8 != 2) continue;

        final raw = toDouble(
          event[i * 3 + 0],
          event[i * 3 + 1],
          event[i * 3 + 2],
        );

        final output = _adaptiveHR.calculateHR(
          raw,
          true,
        );

        if (output > 0) yield output;
      }
    }
  }

  Stream<Channels> _getBufferedStream() async* {
    final events = streams.recentStream;
    if (events == null) return;

    final res = List.generate(12, (index) => <double>[]);

    await for (final event in events) {
      for (var index = 0; index < 9; index++) {
        final l1 = event[index * 8 + 0];
        final l2 = event[index * 8 + 1];
        res[0].add(l1); // L1
        res[1].add(l2); // L2
        res[2].add(l2 - l1); // L3

        res[3].add(-(l1 + l2) / 2); // aVR
        res[4].add(l1 - l2 / 2); // aVL
        res[5].add(l2 - l1 / 2); // aVF

        res[6].add(event[index * 8 + 2]); // V1
        res[7].add(event[index * 8 + 3]); // V2
        res[8].add(event[index * 8 + 4]); // V3
        res[9].add(event[index * 8 + 5]); // V4
        res[10].add(event[index * 8 + 6]); // V5
        res[11].add(event[index * 8 + 7]); // V6
      }

      for (final channel in res) {
        if (channel.length > maxBufferStreamEcg) {
          channel.removeRange(0, channel.length - lengthBufferStreamEcg);
        }
      }
      yield Channels.fromList(res);
    }
  }

  Stream<PlotChannels> _getBufferedPlotStream() async* {
    final events = streams.recentStream;
    if (events == null) return;

    final res = List.generate(12, (index) => <FlSpot>[]);

    final f = isFilteredEcg
        ? (int index, double value) {
            return _dataFilter.applyFilter(
              index,
              value,
              true,
            );
          }
        : (int index, double value) {
            return value;
          };

    var x = 0.0;
    await for (final event in events) {
      for (var index = 0; index < 9; index++) {
        final l1 = f(0, event[index * 8 + 0]);
        final l2 = f(1, event[index * 8 + 1]);
        res[0].add(FlSpot(x, l1)); // L1
        res[1].add(FlSpot(x, l2)); // L2
        res[2].add(FlSpot(x, l2 - l1)); // L3

        res[3].add(FlSpot(x, -(l1 + l2) / 2)); // aVR
        res[4].add(FlSpot(x, l1 - l2 / 2)); // aVL
        res[5].add(FlSpot(x, l2 - l1 / 2)); // aVF

        res[6].add(FlSpot(x, f(2, event[index * 8 + 2]))); // V1
        res[7].add(FlSpot(x, f(3, event[index * 8 + 3]))); // V2
        res[8].add(FlSpot(x, f(4, event[index * 8 + 4]))); // V3
        res[9].add(FlSpot(x, f(5, event[index * 8 + 5]))); // V4
        res[10].add(FlSpot(x, f(6, event[index * 8 + 6]))); // V5
        res[11].add(FlSpot(x, f(7, event[index * 8 + 7]))); // V6
      }

      for (final channel in res) {
        if (channel.length > maxBufferStreamEcg) {
          channel.removeRange(0, channel.length - lengthBufferStreamEcg);
        }
      }

      yield PlotChannels.fromList(res, 0);
      x += 9;
    }
  }

  Stream<PlotChannels> _getBufferedPlotRTStream() async* {
    final events = streams.recentStream;
    if (events == null) return;

    final res = List.generate(12, (index) => <FlSpot>[]);

    final f = isFilteredEcg
        ? (int index, double value) {
            return _dataFilter.applyFilter(
              index,
              value,
              true,
            );
          }
        : (int index, double value) {
            return value;
          };

    final g = skipSampleCount == 1
        ? (List<List<FlSpot>> result) {
            return result;
          }
        : (List<List<FlSpot>> result) {
            return result
                .map(
                  (e) => e
                      .where(
                        (item) => item.x.toInt() % skipSampleCount == 0,
                      )
                      .toList(),
                )
                .toList();
          };

    var x = 0;
    await for (final event in events) {
      for (var index = 0; index < 9; index++) {
        final l1 = f(0, event[index * 8 + 0]);
        final l2 = f(1, event[index * 8 + 1]);
        final xd = x.toDouble();

        if (res[0].length < screenBufferStreamingEcg) {
          res[0].add(FlSpot(xd, l1)); // L1
          res[1].add(FlSpot(xd, l2)); // L2
          res[2].add(FlSpot(xd, l2 - l1)); // L3

          res[3].add(FlSpot(xd, -(l1 + l2) / 2)); // aVR
          res[4].add(FlSpot(xd, l1 - l2 / 2)); // aVL
          res[5].add(FlSpot(xd, l2 - l1 / 2)); // aVF

          res[6].add(FlSpot(xd, f(2, event[index * 8 + 2]))); // V1
          res[7].add(FlSpot(xd, f(3, event[index * 8 + 3]))); // V2
          res[8].add(FlSpot(xd, f(4, event[index * 8 + 4]))); // V3
          res[9].add(FlSpot(xd, f(5, event[index * 8 + 5]))); // V4
          res[10].add(FlSpot(xd, f(6, event[index * 8 + 6]))); // V5
          res[11].add(FlSpot(xd, f(7, event[index * 8 + 7]))); // V6
        } else {
          res[0][x % screenBufferStreamingEcg] = FlSpot(
            xd % screenBufferStreamingEcg,
            l1,
          ); // L1
          res[1][x % screenBufferStreamingEcg] = FlSpot(
            xd % screenBufferStreamingEcg,
            l2,
          ); // L2
          res[2][x % screenBufferStreamingEcg] = FlSpot(
            xd % screenBufferStreamingEcg,
            l2 - l1,
          ); // L3

          res[3][x % screenBufferStreamingEcg] = FlSpot(
            xd % screenBufferStreamingEcg,
            -(l1 + l2) / 2,
          ); // aVR
          res[4][x % screenBufferStreamingEcg] = FlSpot(
            xd % screenBufferStreamingEcg,
            l1 - l2 / 2,
          ); // aVL
          res[5][x % screenBufferStreamingEcg] = FlSpot(
            xd % screenBufferStreamingEcg,
            l2 - l1 / 2,
          ); // aVF

          res[6][x % screenBufferStreamingEcg] = FlSpot(
            xd % screenBufferStreamingEcg,
            f(2, event[index * 8 + 2]),
          ); // V1
          res[7][x % screenBufferStreamingEcg] = FlSpot(
            xd % screenBufferStreamingEcg,
            f(3, event[index * 8 + 3]),
          ); // V2
          res[8][x % screenBufferStreamingEcg] = FlSpot(
            xd % screenBufferStreamingEcg,
            f(4, event[index * 8 + 4]),
          ); // V3
          res[9][x % screenBufferStreamingEcg] = FlSpot(
            xd % screenBufferStreamingEcg,
            f(5, event[index * 8 + 5]),
          ); // V4
          res[10][x % screenBufferStreamingEcg] = FlSpot(
            xd % screenBufferStreamingEcg,
            f(6, event[index * 8 + 6]),
          ); // V5
          res[11][x % screenBufferStreamingEcg] = FlSpot(
            xd % screenBufferStreamingEcg,
            f(7, event[index * 8 + 7]),
          ); // V6
        }
        x += 1;
      }
      yield PlotChannels.fromList(
        // await compute(g, res),
        g(res),
        x.toDouble() % screenBufferStreamingEcg,
      );
    }
  }
}
